#!/usr/bin/env python
#-*- coding: ISO-8859-1 -*-
#pylint: disable-msg=E1101,C0103,R0902,R0903
#Author: Valentin Kuznetsov

"""
ReleaseManager is responsible for setting up CMSSW area and run cmsRun job 
to extract run/events from given LFN/PFN list
"""

import os
import re
import stat
import time
import types
from subprocess import Popen, PIPE
import tempfile
import threading
import ConfigParser

import DBSInteraction
from ConfiguredObject import ConfiguredObject
from ThreadPool import ThreadPool
from Utils import natsort, uniqueList, iparser, edmconfig, walktree, cleanup
from Utils import cmsrun_script

class CMSRunManager(ConfiguredObject):
    """Manage cmsRun jobs"""

#    def __init__(self, area, release, config):
    def __init__(self, area, release, config, lfnlist):
        super(CMSRunManager, self).__init__()
        self.area = area         # CMSSW release area to be used, setup by scram
        self.release = release   # CMSSW release to be used
        self.config  = config    # cmsRun configuration
        self.lfnlist = lfnlist   # LFNs to be transferred for cmsRun processing
        self.process = None      # cmsRun process Popen4 object
        self.tmpfilename = None  # cmsRun configuration file name
        self.scriptname = None   # cmsRun job script file name
        self.release_to_use = '' # CMSSW released used to produce output file
        self.outfilename    = '' # output file name
        self.inputparams = None  # input parameters: dataset, eventset

        # find list of available release on our system
        cmd = "scram list -c CMSSW | awk '{print $2}'"
	try:
	    pipe = Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE, close_fds=True)
	    stdout = pipe.stdout.read()
            self.releases = natsort(list(stdout.split()))
	except:
	    traceback.print_exc()
	    self.releases = None
       
        release_re = re.compile("CMSSW_([0-9]+)_([0-9]+)_([0-9]+)(.*)")
        def cmp2(rel1, rel2):
            m1 = release_re.match(rel1)
            m2 = release_re.match(rel2)
            if  not m1 or not m2:
                return 0
            maj1, min1, fix1, other1 = m1.groups()
            maj1, min1, fix1 = int(maj1), int(min1), int(fix1)
            maj2, min2, fix2, other2 = m2.groups()
            maj2, min2, fix2 = int(maj2), int(min2), int(fix2)
            if maj1 == maj2 and min1 == min2 and fix1 == fix2 and other1 == \
                    other2:
                return 0
            if maj1 < maj2 or (maj1 == maj2 and min1 < min2) or (maj1 == maj2 \
                    and min1 == min2 and fix1 < fix2) or (maj1 == maj2 and \
                    min1 == min2 and fix1 == fix2 and other2 == "") or \
                    (maj1 == maj2 and min1 == min2 and fix1 == fix2 and \
                    other1 < other2):
                return -1
            return 1

        self.releases.sort(cmp=cmp2)
        # find best matched release wrt one we need to use
        for rel in self.releases:
            if  not cmp2(rel, self.release):
                self.release_to_use = rel
		if  rel == self.release:
		    break

    def setup_cmssw(self):
        """setup CMSSW area and run scram"""
        print("Setting up cmssw for %s in %s" % (self.release_to_use, self.area))
        cdir = os.getcwd()
        os.chdir(self.area)
        cmd = 'scramv1 project CMSSW %s' % self.release_to_use
        self.log.info(cmd)
#        print "we are in", os.getcwd()
#        print "execute", cmd
        os.system(cmd)
#        print "go to", cdir
        os.chdir(cdir)
#        print "done setup_cmssw"

    def start(self):
        """start cmsRun job"""
        self.log.info("Starting the cmsRun job")
#        print "starting the cmsRun job"
#        print "release", self.release_to_use, self.release
        cdir = os.getcwd()
        self.setup_cmssw()
        os.chdir( os.path.join( self.area, self.release_to_use ))
        if  self.release_to_use < 'CMSSW_1_6':
            self.log.error("No support for release %s" % self.release_to_use)
            return # no support
        elif self.release_to_use < 'CMSSW_2_1':
            tmpfilename = tempfile.mktemp() + '.cfg'
        else:
            tmpfilename = tempfile.mktemp() + '.py'
        self.tmpfilename = tmpfilename
        tfile = open(tmpfilename, 'w')
        tfile.write(self.config)
        tfile.close()
        self.log.info('Config file %s' % tmpfilename)
        # generate execution script
        script = cmsrun_script(self.lfnlist, tmpfilename)
        scriptname = tempfile.mktemp() + '.sh'
        sfile = open(scriptname, 'w')
        sfile.write(script)
        sfile.close()
	os.chmod(scriptname, 0755)
	self.scriptname = scriptname
#        print "script file", scriptname
#        cmd = 'eval `scramv1 run -sh` && cmsRun %s' % tmpfilename
#        print "Current location", os.getcwd()
#        print "Generated execution script\n", script
        self.log.info('cmsRun execution script\n %s' % script)
	try:
	    self.process = Popen(scriptname, shell=True, stdout=PIPE, stderr=PIPE, close_fds=True)
	except:
	    traceback.print_exc()
	    self.process = None
        os.chdir(cdir)
        
    def process_output(self):
        """Read content of running process"""
        if  self.process:
	    try:
                self.process.wait()
	        return self.process.stdout.read()
	    except:
	        return "ERROR"
        return
        
    def cancel(self):
        """cancel cmsRun job"""
        # TODO: implement kill -9
        if  self.process:
            pid = self.process.pid
            os.system('kill -9 %s' % pid)
            self.process = None
            os.remove(self.tmpfilename)
        return

    def status(self):
        """check status of cmsRun job"""
        status = None
        if  self.process:
	    try:
                status = self.process.poll()
	    except:
		traceback.print_exc()
		pass
            if  status != None: # process.poll() == None means we still doing job
		if  status < 0: # -N indicates that the child was terminated by signal N
		    (child_stdout, child_stderr) = (self.process.stdout, self.process.stderr)
		    self.log.error("\n### Child stderr\n", child_stderr.read())
		    self.log.error("\n### Child stdout\n", child_stdout.read())
                self.process = None
                os.remove(self.tmpfilename)
		os.remove(self.scriptname)
        return status

    def filecontent(self):
        """Read content of output file using emdFileUtil"""
        cwd  = os.getcwd()
        area = os.path.join(self.area, self.release_to_use)
        os.chdir(area)
        cmd  = 'eval `scramv1 run -sh` && edmFileUtil -e -f file:%s' \
                % self.outfilename
	try:
	   pipe = Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE, close_fds=True)
	   stdout = pipe.stdout.read()
	   content = stdout
	except:
	   traceback.print_exc()
	   content = "ERROR2"

        os.chdir(cwd)
        return content
    
    def nevents(self):
        """Use ROOT to check how many events are written in output file"""
        cdir = os.getcwd()
        tmpfilename = tempfile.mktemp() + '.C'
        template = """
using namespace std;

void %s() {
    TFile *f = TFile::Open("%s");
    TTree* tree;
    if ( tree = (TTree*)f->Get("Events") ) {
        cout << tree->GetEntries() << endl;
    }
}
    """ % (tmpfilename.split('/')[-1].replace('.C',''), self.outfilename)
        tfile = open(tmpfilename, 'w')
        tfile.write(template)
        tfile.close()
        area = os.path.join(self.area, self.release_to_use)
        os.chdir(area)
        cmd = 'eval `scramv1 run -sh` && root -l -n -q -b %s 2> /dev/null ' \
            '| tail -1' % tmpfilename
	try:
	    pipe = Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE, close_fds=True)
	    stdout = pipe.stdout.read()
	    content = stdout
	except:
	    traceback.print_exc()
	    content= "ERROR3"

        if  content.find('Processing') != -1:
            content = 0
        os.remove(tmpfilename)
        os.chdir(cdir)
        return content

class CMSRunThreadWrapper(ConfiguredObject):
    """
       this class handles setup of CMSSW area and running cmsRun job
    """
    def __init__(self, area, cp=None):

        self.cp           = None
        self.section      = "release_manager"
        super(CMSRunThreadWrapper, self).__init__(area)
        self.dbs = DBSInteraction.DBS('cms_dbs_prod_global')
        self._request     = {}
        self.max_cmsrun   = 3
        self.request_lock = threading.Lock()
        self.pool         = None

    def configure(self, cp):
        """configure wrapper"""
        self.cp = cp
        self.max_cmsrun = int(self.getOption("max_cmsrun", "5"))
        self.pool = ThreadPool(threads=self.max_cmsrun)
        
    def submit(self, user, requestdict):
        """
        Submit a cmsRun job over a provided request dictionary.
        """
        eventlist = requestdict.get('eventlist', [])
        dataset = requestdict.get('dataset', 'NONE')
        site = requestdict.get('site', None)
        if not dataset:
            raise Exception("Request dictionary did not include dataset.")
        
        # Calculate the LFNs needed for this request.
	res  = list()
        for run, _, lumi in eventlist:
	    files = self.dbs.getFiles(dataset=dataset, run=run, lumi=lumi, 
                        site=site)
	    if  files not in res:
		for item in files:
	            res.append(item)
        result = list(res)
        if not result:
            raise Exception("No files found at site=%s" % site)

        cmsfs_location = None
        if self.cp:
            cmsfs_location = self.getOption("cmsfs_location", None)
        
        # Calculate the releases needed to process request.
        rellist = []
        lfnlist = []
        for lfn, _ in result:
	    if  type(lfn) is types.UnicodeType:
		lfnlist.append(unicode.encode(lfn))
	    else:
                lfnlist.append(lfn)
            release = self.dbs.getRelease(lfn)
            if  not rellist.count(release):
                rellist.append(release)
        rellist.sort()

        # check that release list has 1 value
        if  len(rellist) > 1:
            msg  = 'Ambiguous request.\n'
            msg += 'Found multiple CMSSW releases for input parameter set.'
            raise Exception(msg)
        elif len(rellist) == 0:
            raise Exception("No release found for request.")
        
        release  = rellist[0]
        cmsswdir = self.getOption("cmssw_dir", "/tmp/cms_fs_files")
        odir     = os.path.join(cmsswdir, 'root_files')
        opath    = os.path.join(odir, str(user))
        try:
            os.makedirs(opath)
        except:
            pass
        walktree(odir, cleanup) # clean-up top-level dir
        outfilename = os.path.join(opath, \
                                   '%s.root' % str(time.time()).replace('.',''))

        if  os.path.isfile(outfilename):
	    print "\n### Remove existing file", outfilename
            os.remove(outfilename)

        if cmsfs_location:
            lfnlist = ['file:' + cmsfs_location + '/lfns' + i for i in lfnlist]

#        config = edmconfig(release, lfnlist, eventlist, outfilename)
        config = edmconfig(release, lfnlist, eventlist, outfilename, prefix='file:////tmp')
        if  not config:
            msg  = 'We unable to generate EDM config file based on provided ' \
                'information.<br/>'
            raise Exception(msg)
#        requestId = self.request(cmsswdir, release, config)
        requestId = self.request(cmsswdir, release, config, lfnlist)
        if requestId:
            self.updateparams(requestId, outfilename, requestdict)
        return requestId

#    def request(self, area, release, config):
    def request(self, area, release, config, lfnlist):
        """
        Run cmsRun job over provided config file.
        Returns a requestId which can be used later for status checks.
        """
        if  len(self._request.keys()) == self.max_cmsrun:
            return None # to prevent from having too many cmsrun jobs on a node
        requestid = str(time.time()).replace('.','')
        self.request_lock.acquire()
        try:
            try:
#                cmsrun = CMSRunManager(area, release, config)
                cmsrun = CMSRunManager(area, release, config, lfnlist)
                self._request[requestid] = cmsrun
                self.pool.queue(cmsrun)
            except RuntimeError, e:
                # unable to start new thread
                self.log.exception(e)
                requestid = None
        finally:
            self.request_lock.release()
        return requestid

    def status(self, requestid):
        """
           return status of cmsRun job
        """
        status = -1
        if  not requestid:
            return 'busy'
        self.request_lock.acquire()
        try:
            try:
                cmsrun = self._request[requestid]
                status = cmsrun.status()
                return status
            except (SystemExit, KeyboardInterrupt):
		traceback.print_exc()
                raise
            except Exception, e:
		traceback.print_exc()
                self.log.exception(e)
                return status
        finally:
            self.request_lock.release()
        return status
        
    def updateparams(self, requestid, outfilename, params):
        """update object with given input parameters for given requestid"""
        self._request[requestid].outfilename = outfilename
        self._request[requestid].inputparams = params
        # look-up for old files and clean them up
        # delete files older then 3 hours
        odir = os.path.split(outfilename)[0]
        for ifile in os.listdir(odir):
            ofile = os.path.join(odir, ifile)
            filestat = os.stat(ofile)
            if  filestat[stat.ST_MTIME] > (60*60*24):
                os.remove(ofile)
        
    def inputparams(self, requestid):
        """return input parameters for given requestid"""
        return self._request[requestid].inputparams
        
    def getoutfile(self, requestid):
        """return output file name for given requestid"""
        return self._request[requestid].outfilename
        
    def filecontent(self, requestid):
        """return file content of requestid"""
        return self._request[requestid].filecontent()
        
    def nevents(self, requestid):
        """return number of events in output file corresponding to requestid"""
        return self._request[requestid].nevents()

    def release_to_use(self, requestid):
        """return file content of requestid"""
        return self._request[requestid].release_to_use
        
    def config(self, requestid):
        """return file content of requestid"""
        return self._request[requestid].config

    def process_output(self, requestid):
        """return output of cmsRun for requestid"""
        return self._request[requestid].process_output()
        
    def clean(self, requestid):
        """clean internal cache about requestid"""
	try:
            del self._request[requestid]
	except:
	    traceback.print_exc()
        
    def cancel(self, requestid):
        """cancel requested cmsRun job"""
        self.request_lock.acquire()
        try:
            cmsrun = self._request[requestid]
            self.log.info("Sending a cancel request to cmsrun request=%s, " \
                          "filename=%s" % (requestid, cmsrun.outfilename))
            cmsrun.cancel()
            self.clean(requestid)
	except:
	    traceback.print_exc()
        finally:
            self.request_lock.release()

    def graceful_exit(self):
        """implement exit from thread pool"""
        self.pool.drain()
        self.pool.join()
#
# main
#
def test(dbsmgr, mgr, inputdir, dataset, eventset):
    """test request"""
    eventlist = [x for x in iparser(eventset)]
    result = []
    for run, evt, lumi in eventlist:
        for item in dbsmgr.getFiles(run, dataset, evt, lumi):
	    if  item not in result:
	        result.append(item)
    lfnlist = []
    rekease = None
    for item in result:
	lfn, size = item
        release = dbsmgr.getRelease(lfn)
	if  lfn not in lfnlist:
            lfnlist.append(lfn)
    outfilename = inputdir + '/root_files/valya/test.root'
    config = edmconfig(release, lfnlist, eventlist, outfilename, prefix='file:////tmp')
    requestid = mgr.request(inputdir, release, config, lfnlist)
    if  not requestid:
        print "#### UNABLE to start thread for", dataset, eventset
    return requestid

def check_status(mgr, requestid):
    """test status"""
    from Utils import sendEmail
    counter = 0
    outfilename = mgr.getoutfile(requestid)
    while 1:
        status = mgr.status(requestid)
        if  status == 'busy':
            break
        if  os.path.isfile(outfilename):
            print '%s %s' % (os.stat(outfilename)[stat.ST_SIZE], outfilename)
        else:
            print 'cmsRun job is launched, please wait, job id: %s, status %s'\
                 % (requestid, status)
        if  status != None:
	    msg = "Finish job %s status %s" % (requestid, status)
	    print msg
	    sendEmail('vkuznet@gmail.com', msg, requestid)
            break
        counter += 1
        time.sleep(5)

def main():
    """main test"""
    inputdir = '/tmp/cms_fs_files'
    mgr = CMSRunThreadWrapper(inputdir)
    cp = ConfigParser.ConfigParser()
    cp.add_section("release_manager")
    cp.set("release_manager", "max_cmsrun", "2")
    mgr.configure(cp)
    import thread
    import DBSInteraction
    dbsmgr = DBSInteraction.DBS('cms_dbs_prod_global')
    # test CMSSW_1_6_X release series
#    dataset = '/testbeam_HCalEcalCombined/h2tb2007_default_v1/DIGI-RECO'
#    eventset = "16278 44000 10"
#    requestid = test(dbsmgr, mgr, inputdir, dataset, eventset)
#    print "####", requestid, mgr.getoutfile(requestid)
#    if  requestid:
#        thread.start_new_thread(check_status, (mgr, requestid))
    # test CMSSW_2_1_X release series
#    dataset  = '/Cosmics/Commissioning08-CRUZET4_v1/RECO'
#    eventset = "58620 151578 6" 
#    requestid = test(dbsmgr, mgr, inputdir, dataset, eventset)
#    print "####", requestid, mgr.getoutfile(requestid)
#    if  requestid:
#        thread.start_new_thread(check_status, (mgr, requestid))
    # use the same dataset but real event, this time we should fail since 
    # we only allow 2 cmsRun jobs
#    eventset = "58620 151578 10" 
#    requestid = test(dbsmgr, mgr, inputdir, dataset, eventset)
#    print "####", requestid, mgr.getoutfile(requestid)
#    if  requestid:
#        thread.start_new_thread(check_status, (mgr, requestid))
    # test CMSSW_3_5_6 release
    dataset = '/MinimumBias/Commissioning10-PromptReco-v8/RECO'
    eventset = '132601 1333447 154'
    requestid = test(dbsmgr, mgr, inputdir, dataset, eventset)
    if  requestid:
        thread.start_new_thread(check_status, (mgr, requestid))

    time.sleep(180)

if __name__ == "__main__":
    main()
