
"""
The GridFTP-block portion of the CmsFs filesystem.
"""

import os
import time
import logging
import threading

from pyGlobus.ftpClient import FtpClient, HandleAttr #pylint: disable-msg=F0401
from pyGlobus.ftpClient import OperationAttr #pylint: disable-msg=F0401
from pyGlobus.ftpClient import RestartMarker #pylint: disable-msg=F0401
from pyGlobus.util import Buffer, GlobusException #pylint: disable-msg=F0401

from ConfiguredObject import ConfiguredObject

from CmsFsTransfers import Block
from FileLookup import LookupServer
from FileManager import SimpleCron

log = logging.getLogger("CmsFs.Cache")

# The following connection cache/cleaner c code allows the client to re-use
# GridFTP connections.  However, the Globus C code crashes and I can't figure
# out why quite yet ... therefore, it's commented out for now.
_conn_lock = threading.Lock()
_conn_caches = {}
_handles = []
_clients = []

def conn_cleaner():
    _conn_lock.acquire()
    delete_list = []
    try:
        for turl, val in _conn_caches.items():
            timestamp, gftp = val
            if time.time() - timestamp > 60: 
                #gftp.flush_url_state(turl)
                delete_list.append(turl)
        for turl in delete_list:
            del _conn_caches[turl]
    finally:
        _conn_lock.release()

#ConnectionCleaner = SimpleCron("Connection Cache cleaner thread",
#    conn_cleaner, 90)

class GridftpPartialRunner(ConfiguredObject):

    """
    This class is responsible for filling up partial blocks based upon the
    request and the current contents of the partial_file.

    It is responsible for resolving the LFN to a SURL, then the SURL to a TURL.
    Once this is done, it performs the partial GridFTP transfer
    """

    def __init__(self, block_request, partial_file):
        super(GridftpPartialRunner, self).__init__()
        self.block = partial_file.calculate_partial_block(block_request)
        block_store = partial_file.block_store
        self.block_request = block_request
        self._condExternal = threading.Condition(threading.Lock())
        if isinstance(self.block, Block): # We have a complete block on disk.
            # We don't need to serialize this access, as only one thread can be
            # in the constructor at a time.
            if not self.block.data_loaded():
                block_store.retrieve(self.block, partial_file)
                self.log.debug("GridftpRunner can be fulfilled by an existing" \
                    " block which we just retrieved.")
            else:
                self.log.debug("GridftpRunner can be fulfilled by an existing "\
                    "block.")
            self.can_satisfy = True
            log.info("HIT: %s, %s" % (partial_file, block_request))
            return
        self.log.debug("Starting GridftpRunner for %s." % self.block_request)
        if block_request.partial_block_can_satisfy(self.block):
            self.log.debug("Trivial!  Can already be satisfied.")
            self.can_satisfy = True
            log.info("HIT: %s, %s" % (partial_file, block_request))
            return
        log.info("MISS: %s, %s" % (partial_file, block_request))
        self.can_satisfy = False
        self._buffer = Buffer(256*1024)
        self.handleAttr = HandleAttr()
        #self._ftpClient = None # Commented-out handle cache code
        self._ftpClient = FtpClient(self.handleAttr)
        self._condV = threading.Condition(threading.Lock())
        self.cp = partial_file.cp
        self.partial_file = partial_file
        self.error = None
        self.handle = None
        self.src_url = None

    def wait_on_data(self, timeout=None):
        """
        This function blocks until the block_request is fulfilled or
        there is some GridFTP error.
        """
        if (isinstance(self.block, Block) and self.block_request.\
                block_can_satisfy(self.block)) or self.block_request.\
                partial_block_can_satisfy(self.block):
            return self.block.fulfill_request(self.block_request)
        self._condExternal.acquire()
        self._condExternal.wait(timeout=timeout)
        self._condExternal.release()
        if self.error:
            raise self.error #pylint: disable-msg=W0706
        elif self.block_request.partial_block_can_satisfy(self.block):
            return self.block.fulfill_request(self.block_request)
        elif timeout:
            return None
        else:
            #raise Exception("Improperly signaled completion")
            return self.block.data[self.block_request.offset - \
                self.block.offset:]

    def read_cb(self, arg, handle, buf, bufHandle, bufLen, offset, eof,
            error):
        """
        The read callback function from the GridFTP server.
        """
        if error[0] != 0:
            self.error = Exception("GridFTP error: %s" % str(error))
            self.log.error(self.error)
            self._condExternal.acquire()
            self._condExternal.notifyAll()
            self._condExternal.release()
            raise self.error #pylint: disable-msg=W0706
        self.block.write(buf)
        if self.block_request.partial_block_can_satisfy(self.block):
            self._condExternal.acquire()
            self._condExternal.notifyAll()
            self._condExternal.release()
        if not eof:
            try:
                handle = self._ftpClient.register_read(self._buffer,
                    self.read_cb, None)
                if self.handle != None:
                    self._ftpClient.free_callback(self.handle)
                self.handle = handle
                # Commented-out connection cache code.
                #_handles.append(self.handle)
            except Exception, ex:
                self.error = ex
                self._condExternal.acquire()
                self._condExternal.notifyAll()
                self._condExternal.release()
                self.log.error(self.error)
                raise self.error #pylint: disable-msg=W0706
        else: # We're at EOF
            # TODO: hack to shorten the block request.
            self.log.info("Done with GridFTP read.")
            self.log.debug("Shortening the block request! Requested %s, " \
                "current %s." % (self.block.requested, self.block.current))
            self.block.requested = self.block.current
            self.block.complete = True
            # Connection cache code
            #_conn_lock.acquire()
            #try:
            #    if self.src_url in _conn_caches and _conn_caches[\
            #            self.src_url][0] - time.time() > 30:
            #        pass
            #        #_conn_caches[self.src_url][1].flush_url_state(self.src_url)
            #    if self.src_url in _conn_caches and _conn_caches[\
            #            self.src_url][0] - time.time() <= 30:
            #        pass
            #    else:
            #        self.log.info("\nCaching URL for %s\n" % self.src_url)
            #        _conn_caches[self.src_url] = time.time(), self._ftpClient
            #        #self._ftpClient.cache_url_state(self.src_url)
            #finally:
            #    _conn_lock.release()
            self._condExternal.acquire()
            self._condExternal.notifyAll()
            self._condExternal.release()

    def done_func(self, cv, _, error):
        """
        Cleanup at end of the GridFTP transfer.
        """
        if error[0] != 0:
            self.log.error("The following error occured: %s" % str(error))
            self.error = Exception("GridFTP error: %s" % str(error))
            raise self.error #pylint: disable-msg=W0706
        else:
            self.log.info("No Errors")
            self.block.finish()
        cv.acquire()
        cv.notifyAll()
        cv.release()
        if error[0] != 0:
            raise self.error #pylint: disable-msg=W0706
        if not self.src_url:
            return
        # Connection cache code.
        #_conn_lock.acquire()
        #try:
        #    if self.src_url in _conn_caches and _conn_caches[\
        #           self.src_url][0] - time.time() > 30:
        #        _conn_caches[self.src_url][1].flush_url_state(self.src_url)
        #    elif self.src_url in _conn_caches and _conn_caches[\
        #            self.src_url][0] - time.time() <= 30:
        #        pass
        #    else:
        #        print "\nCaching URL for %s\n" % self.src_url
        #        _conn_caches[self.src_url] = time.time(), self._ftpClient
        #        self._ftpClient.cache_url_state(self.src_url)
        #finally:
        #    _conn_lock.release()
        #self._ftpClient.cache_url_state(self.src_url)


    def start(self):
        """
        Start the GridFTP transfer.
        Actually, this is a wrapper around the actual start function in order
        to make sure that any exception is caught and any consumers are
        notified of the error.
        """
        try:
            self._start_real()
        except Exception, error:
            self.error = Exception("GridFTP error: %s" % str(error))
            self._condExternal.acquire()
            self._condExternal.notifyAll()
            self._condExternal.release()
            raise

    def _start_real(self):
        """
        The actual implementation of the start of the GridFTP transfer.
        """
        if (isinstance(self.block, Block) and self.block_request.\
                 block_can_satisfy(self.block)) or self.block_request.\
                 partial_block_can_satisfy(self.block):
            return
        lfn = self.block.parent_file.lfn
        self.log.info("Start Gridftp transfer for LFN %s." % lfn)
        pfn = self.partial_file.getPfn()
        self.log.info("Returned PFN %s for LFN %s." % (pfn, lfn))
        src_url = LookupServer.acquireTURL(pfn)
        self.log.info("Got TURL %s for PFN %s." % (pfn, src_url))
        self.src_url = src_url
        # Commented-out connection cache code
        #_conn_lock.acquire()
        #try:
        #    if self.src_url in _conn_caches and time.time() - \
        #            _conn_caches[self.src_url][0] <= 60:
        #        self._ftpClient = _conn_caches[self.src_url][1]
        #        self.log.info("\nUsing cached ftp client for %s\n" % \
        #            self.src_url)
        #        del _conn_caches[self.src_url]
        #    else:
        #        self.log.info("\nStarting new ftp client for %s\n" % \
        #            self.src_url)
        #        self._ftpClient = FtpClient(self.handleAttr)
        #        _clients.append(self._ftpClient)
        #finally:
        #    _conn_lock.release()
        self.log.info("Returned TURL %s for PFN %s." % (src_url, pfn))
        try:
            operationAttr = OperationAttr()
            restartMarker = RestartMarker()
            self.log.info("Offset %i, length %i" % (self.block.offset,
                self.block.requested))
            self._ftpClient.partial_get(src_url, operationAttr, restartMarker,
                self.block.offset, self.block.offset + self.block.requested,
                self.done_func, self._condV)
            self.handle = self._ftpClient.register_read(self._buffer,
                self.read_cb, None)
            #_handles.append(self.handle) # Commented-out connection cache code
            #print "Done register_read"
        except GlobusException, _:
            if self.src_url:
                LookupServer.releaseTURL(self.src_url)
                self.src_url = None
            raise

        self._condV.acquire()
        self._condV.wait()
        self._condV.release()
        #fd.close()
        if self.src_url:
            LookupServer.releaseTURL(self.src_url)
            self.src_url = None
        self.log.info("Finished Gridftp read for %s cleanly." % src_url)
        if self.handle:
            self._ftpClient.free_callback(self.handle)

    def cancel(self):
        """
        Cancel an in-progress GridFTP transfer
        """
        if self.src_url:
            LookupServer.releaseTURL(self.src_url)
            self.src_url = None
        if not self.block.complete:
            self._ftpClient.abort()


