
import os
import sets
import array
import shelve
import threading

from ConfiguredObject import ConfiguredObject
from FileLookup import LookupServer

class BlockRequest(object):

    def __init__(self, offset, requested):
        self.offset, self.requested = offset, requested

    def block_can_satisfy(self, block):
        if block.offset <= self.offset and block.offset + block.requested >= \
                self.offset + self.requested:
            return True

    def partial_block_can_satisfy(self, partial_block):
        if partial_block.offset <= self.offset and partial_block.offset + \
                partial_block.current >= self.offset + self.requested:
            return True
        return False

    def transfer_can_satisfy(self, transfer):
        if transfer.current_marker >= self.offset + self.requested:
            return True

    def __str__(self):
        return "BlockRequest<%i,%i>" % (self.offset, self.requested)

class Block(object):
 
    """
    Object representing a single block.

    This serializes to a basic (offset, requested) tuple.  However, if given
    a block store, it can load block data into memory.

    This is not thread-safe, and usage must be synchronized externally.

    To access data in this interface, you must do something like this:

    if not blk.data_loaded():
        block_store.retrieve(block, partial_file) # Loads into memory
    blk.fulfill_request(block_request) # Returns actual data.

    The first two lines should be done with a lock if multiple threads might be
    accessing it.
    """

    def __init__(self, offset, requested):
        self.offset = offset
        self.requested = requested
        self.data = None

    def data_loaded(self):
        return self.data != None

    def fulfill_request(self, block_request):
        if self.data == None:
            raise Exception("Attempted to fulfill request before data was set.")
        my_start = block_request.offset - self.offset
        my_finish = block_request.offset + block_request.requested - self.offset
        return self.data[my_start:my_finish]

    def merge(self, block):
        raise NotImplementedError()

    def __getstate__(self):
        """
        The state of this object is only dependent on the offset and length
        """
        return (self.offset, self.requested)

    def __setstate__(self, state):
        self.data = None
        self.offset, self.requested = state

    def __eq__(self, other):
        if not isinstance(other, Block):
            return False
        return self.__hash__() == other.__hash__()

    def __hash__(self):
        return (self.offset, self.requested).__hash__()

class PartialBlock(ConfiguredObject):

    """
    Object representing a partial, in-memory block.  When this PartialBlock
    is complete, it will be stored in the BlockStore
    """

    def __init__(self, offset, requested, parent_file, data=None):
        super(PartialBlock, self).__init__()
        self.offset, self.requested = offset, requested
        self.parent_file = parent_file
        if data:
            assert len(data) == requested
            self.data = data
            self.complete = True
            self.current = requested
        else:
            self.data = array.array('c')
            self.current = 0
            self.complete = False

    def __str__(self):
        return "PartialBlock<offset %i,requested %i,completed %i>" % \
            (self.offset, self.requested, self.current)

    def write(self, data):
        self.data.extend(data)
        self.current += len(data)
        self.log.debug("Write to PartialBlock: %s" % self)

    def fulfill_request(self, block_request):
        my_start = block_request.offset - self.offset
        my_finish = block_request.offset + block_request.requested - self.offset
        self.log.debug("%s satisfied the request %s" % (self, block_request))
        return self.data[my_start:my_finish]

    def merge(self, partial_block):
        assert self.complete == True and partial_block.complete == True
        if partial_block.offset == self.offset:
            if partial_block.requested >= self.requested:
                return partial_block
            else:
                return self
        if partial_block.offset < self.offset:
            return partial_block.merge(self)
        # self.offset < partial_block.offset
        if self.offset + self.requested >= partial_block.offset:
            if self.offset + self.requested >= partial_block.offset + \
                    partial_block.requested:
                return self
            keep_len = self.offset + self.requested - partial_block.offset
            assert keep_len > 0 and keep_len < self.requested
            new_data = self.data[:keep_len] + partial_block.data
            new_size = keep_len + partial_block.requested
            assert new_size == len(new_data)
            self.log.info("Created a new merged PartialBlock of size %i." % \
                new_size)
            return PartialBlock(self.offset, new_size, self.parent_file,
                data=new_data)
        # Blocks do not overlap or touch.
        return None

    def finish(self):
        self.log.info("%s finished." % (self))
        self.parent_file.finish_partial_block(self)

class PartialFileFactory(ConfiguredObject):

    """
    A factory to produce PartialFile objects.  This allows objects
    to be cached and reused as necessary.
    """

    def __init__(self):
        super(PartialFileFactory, self).__init__()
        self._lock = threading.Lock()
        self.partial_files = {}
        self.cp = None
        self.block_store = None

    def configure(self, cp):
        self.cp = cp

    def getFile(self, guid, lfn, size=-1):
        if not self.block_store:
            raise Exception("Partial File Factory has not been associated with"\
                " a block store.")
        self._lock.acquire()
        try:
            if (guid, lfn) in self.partial_files:
                return self.partial_files[guid, lfn]
            pf = PartialFile(guid, lfn, self.block_store)
            if size >= 0:
                pf.setSize(size)
            self.log.info("Creating a new PartialFile for %s" % \
                str((guid, lfn)))
            pf.configure(self.cp)
            return self.partial_files.setdefault((guid, lfn), pf)
        finally:
            self._lock.release()

    def __call__(self, guid, lfn, size=-1):
        return self.getFile(guid, lfn, size=size)

class PartialFile(ConfiguredObject):

    """
    Class representing a PartialFile.
    Each partial file has both blocks and partial blocks
    When a partial block is completed, it notifies its PartialFile, which then
    writes is out to the correct blockstore

    Each PartialFile has a unique (lfn, guid) pair.
    """

    def __init__(self, guid, lfn, block_store):
        super(PartialFile, self).__init__()
        self.partial_blocks = []
        self.blocks = block_store.get_blocks(guid, lfn)
        self.block_store = block_store
        self.min_read = 1*1024*1024
        self.guid = guid
        self.lfn = lfn
        self._lock = threading.Lock()
        self.pfn = None
        self.log.info("Initialized object %s." % self)
        self.size = None
        
    def __str__(self):
        return "PartialFile<GUID: %s, LFN: %s, Blocks: %i>" % (self.guid,
            self.lfn, len(self.blocks))

    def configure(self, cp):
        self.cp = cp
        self.file_lookup = LookupServer

    def setSize(self, size):
        self.size = size
        
    def getPfn(self):
         self._lock.acquire()
         try:
             if self.pfn == None:
                 self.pfn = self.file_lookup.getPFN(self.lfn,
                     protocol='srmv2')
                 self.log.info("getPfn result: LFN %s -> PFN %s" % (self.lfn,
                     self.pfn))
             return self.pfn
         finally:
             self._lock.release()

    def calculate_partial_block(self, block_request):
        """
        Create a partial block which could satisfy this request.
        """
        for block in self.blocks:
            if block_request.block_can_satisfy(block):
                self._lock.acquire()
                try:
                    if not block.data_loaded():
                        self.block_store.retrieve(block, self)
                finally:
                    self._lock.release()
                return block
        self._lock.acquire()
        try:
            for block in self.partial_blocks:
                if block_request.partial_block_can_satisfy(block):
                    return block
        finally:
            self._lock.release()
        # TODO: Make sure we don't overlap with any current partial block!
        size = max(block_request.requested, self.min_read)
        if self.size != None and block_request.offset + size > self.size:
            size = self.size - block_request.offset
        self.log.info("Creating PartialBlock of size %i" % size)
        partial_block = PartialBlock(block_request.offset, size, self)
        self._lock.acquire()
        try:
            self.partial_blocks.append(partial_block)
        finally:
            self._lock.release()
        return partial_block
    
    def finish_partial_block(self, partial_block):
        """
        Called when a partial block has been finished off.
        
        This will save it to the file store, and it will become a full block.
        """
        assert partial_block.requested == len(partial_block.data)
        blocks = self.block_store.store(partial_block)
        self._lock.acquire()
        try:
            self.blocks = blocks
            self.partial_blocks.remove(partial_block)
        finally:
            self._lock.release()


class BlockStore(ConfiguredObject):

    """
    The BlockStore is responsible for saving completed blocks to disk.
    """

    def __init__(self, cp, pff):
        super(BlockStore, self).__init__(cp)
        self.cp = cp
        self.partial_file_factory = pff
        pff.block_store = self
        self.section = "block_store"
        self.store_dir = self.getOption("directory", "/tmp/blockstore")
        self.log.info("Initializing the BlockStore in directory %s." % \
            self.store_dir)
        self._lock = threading.Lock()
        self.init_store()

    def get_store_dir(self):
        return self.store_dir

    def init_store(self):
        if not os.path.exists(self.store_dir):
            os.makedirs(self.store_dir)
        try:
            self.shelve = shelve.open(os.path.join(self.store_dir,
                "block_index.db"), writeback=False)
        except Exception, e:
            self.log.error("Unable to load the block store directory.")
            raise
        self.log.debug("Loading block store directory.")
        ctr = 0
        for partial_file, blocks in self.shelve.items():
            self.log.debug("Loading blocks for %s" % str(partial_file))
            ctr += 1
            guid, lfn = eval(partial_file, {}, {})
            # When the partial file is constructed, it will call get_blocks
            # to load up the known blocks.
            # We first check to make sure all blocks are valid.
            ctr2 = 0
            for block in blocks:
                ctr2 += 1
                file = self._getFilename(block, guid=guid)
                if not os.path.exists(file) or os.stat(file).st_size != \
                       block.requested:
                    self.invalidate(block, guid, lfn)
            pf = self.partial_file_factory(guid, lfn)
            self.log.info("Loaded %i blocks for %s." % (ctr2, str(pf)))
        self.log.info("Loaded %i files during BlockStore init." % ctr)

    def store(self, partial_block):
        file = self._getFilename(partial_block)
        self.log.info("Saving a completed block to %s." % file)
        self.log.info("Block is %s." % partial_block)
        fp = open(file, 'w')
        data = partial_block.data
        if isinstance(data, array.array):
            data = data.tostring()
        assert len(data) == partial_block.requested
        fp.write(data)
        fp.close()
        partial_file = partial_block.parent_file
        guid, lfn = partial_file.guid, partial_file.lfn
        self._lock.acquire()
        try:
            key = str((guid, lfn))
            if key not in self.shelve:
                self.shelve[key] = sets.Set()
            set = self.shelve[key]
            set.add(Block(partial_block.offset, partial_block.requested))
            self.shelve[key] = set
            self.shelve.sync()
        finally:
            self._lock.release()
        return set

    def retrieve(self, block, pf):
        """
        Given an instance of Block, retrieve the associated data
        """
        file = self._getFilename(block, partial_file=pf)
        self.log.info("Retrieving a completed block from %s." % file)
        fp = open(file, 'r')
        data = fp.read()
        fp.close()
        block.data = data
        return data

    def invalidate(self, block, guid, lfn):
        """
        Make on-disk data invalid.
        """
        file = self._getFilename(block, guid=guid)
        self.log.info("Invalidating file %s" % file)
        self._lock.acquire()
        try:
            key = str((guid, lfn))
            if key not in self.shelve:
                return
            blockset = self.shelve[key]
            blockset.discard(block)
            print set
            self.shelve[key] = blockset
            self.shelve.sync()
            if os.path.exists(file):
                os.unlink(file)
        finally:
            self._lock.release()

    def get_blocks(self, guid, lfn):
        """
        Returns the blocks this block store knows about for the GUID/LFN.
        """
        self.log.info("Getting blocks for GUID %s, LFN %s." % (guid, lfn))
        results = sets.Set()
        self._lock.acquire()
        try:
            key = str((guid, lfn))
            if key in self.shelve:
                results = self.shelve[key]
            else:
                self.log.info("No block store entries found for %s." % key)
        finally:
            self._lock.release()
        self.log.info("Found %i blocks for GUID %s, LFN %s." % (len(results),
            guid, lfn))
        return results

    def _getFilename(self, block, partial_file=None, guid=None):
        if guid:
            filename = guid + '_' + str(block.offset) + \
                '_' + str(block.requested)
        elif partial_file:
            filename = partial_file.guid + '_' + str(block.offset) + \
                '_' + str(block.requested)
        else:
            filename = block.parent_file.guid + '_' + str(block.offset) + \
                '_' + str(block.requested)
        return os.path.join(self.store_dir, filename)


