#!/usr/bin/env python
#-*- coding: ISO-8859-1 -*-
#
# Author: Brian Bockelman

"""
Manage the running extractions of PickEvent
"""

import sets

import FileManager
import DBSInteraction
import EdmFileExtraction
from Utils import edmconfig
from ConfiguredObject import ConfiguredObject

class EdmFileManager(ConfiguredObject):

    def __init__(self):
        self.section = "edm_file_manager"
        self._server = FileManager.Server
        self._dbs = DBSInteraction.DBS
        self._run_cache = {}
        self._running_extractions = {}
        self._extractions_status = {}

    def configure(self, cp):
        self._server.configure(cp)
        self.pool = FileManager.ThreadPool(threads=int(self.getOption( \
            "max_extractions", "5")))
        self.guid_src = self.getOption("cmsfs_directory", "/tmp/cmsfs/guids")
        self.run_dest = self.getOption("extraction_directory",
            "/tmp/extracted")

    def getEvent(self, run, event):
        raise NotImplementedError()
        # This will be called after edmFileUtil has extracted the event.
        # (status returns OK).  
        
    def _getLfnsForInput(self, dataset, run, lumi, event):
        if (dataset, run, lumi, event) not in self._run_cache:
            self._run_cache[dataset, run, lumi, event] = self._dbs.getFiles( \
                dataset=dataset, run=run, lumi=lumi, event=event)
        return self._run_cache[dataset, run, lumi, event]

    def _handleRunningExtractions():
        failed_keys = []
        self._extraction_lock.acquire()
        try:
            for key, extraction in self._running_extractions.items():
                self._extractions_status[key] = extraction.status()
                if status[0] == 3: # FAILED!
                    failed_keys.append(key)
            for key in failed_keys:
                del self._running_extractions[key]
        finally:
            self._extraction_lock.release()

    def _launchExtraction(self, dataset, event_list):
        key = run, event
        if key in self._running_extractions:
            self.log.warning("Internal error: about to launch an extraction " \
                "for something which is already running! Run %s, event %s" % \
                (run, event))
        extracter = EdmFileExtraction(self.cp)
        extracter.request(run, event, self._getLfnsForRun(run), self.lfn_src,
            self.run_dest)
        self.pool.queue(extracter)
        self._running_extractions[key] = extracter

    def status(self, requestid):
        if requestid in self._running_extractions:
            self._handleRunningExtractions()
        if requestid in self._extractions_status:
            return self._extractions_status[run, event]
        lfns = self._getLfnsForRun(run)
        worst_status = 0
        status_messages = []
        for lfn in lfns:
            status_code, status_msg = self._server.status(lfn)
            if status_code > worst_status:
                worst_status = status_code
            status_messages.append("LFN %s status message: %s" % (lfn,
                status_msg))
        if worst_status > 0:
            return worst_status, '\n'.join(status_messages)
        
        # First, check with the server and report the "worst" file status
        # If all the files are working, return "extraction in progress"
        # This is what launches the edmFileUtil in a separate thread.
        # Once edmFileUtil has successfully found our event, then we
        # return the success code.

    def request(self, dataset, event_list, user=None):
        lfns = sets.Set()
        for run, lumi, event in event_list:
            lfns.update([i[0] for i in self._getLfnsForInput(dataset, run,
                lumi, event)])

    def cancel(self, run, event, user=None):
        for lfn in self._getLfnsForRun(run):
            self._server.cancel(lfn, user=user)

    def graceful_exit(self):
        self._server.graceful_exit()

