
import os
import sys
import sets
import time
import unittest
import threading
import ConfigParser

if os.path.exists("src"):
    sys.path.append("src")
if os.path.exists('../src'):
    sys.path.append('../src')

from CmsFileServer.MappingManager import MappingManager

class BlockingMappingManager(MappingManager):
    """
    A mapping manager which can be made to block on acquire and release
    (so we can test thread-safety a bit).
    """
    
    def __init__(self):
        super(BlockingMappingManager, self).__init__()
        self._acquire_cond = None
        self._release_cond = None
        self._acquire_count = 0
        self._release_count = 0
        self._acquire_error = None
        
    def blockOnAcquire(self, cond):
        self._acquire_cond = cond
        
    def errorOnAcquire(self, error):
        self._acquire_error = error

    def _lookup(self, key):
        if self._acquire_cond:
            self._acquire_cond.acquire()
            self._acquire_cond.wait()
            self._acquire_cond.release()
        self._acquire_count += 1
        if self._acquire_error:
            raise self._acquire_error
        return key**2

    def blockOnRelease(self, cond):
        self._acquire_cond = cond
        
    def _release(self, key):
        if self._release_cond:
            self._release_cond.acquire()
            self._release_cond.wait()
            self._release_cond.release()
        self._release_count += 1
            
class TestMappingManager(unittest.TestCase):
    """
    A test case for the MappingManager
    """
    
    def setUp(self):
        self.manager = BlockingMappingManager()
    
    def test_acquire_release_serial(self):
        value = self.manager.acquireValue(3)
        self.failUnless(value==9, "Returned the incorrect value.")
        self.manager.releaseKey(3)
        value = self.manager.acquireValue(3)
        self.failUnless(value==9, "Returned the incorrect value.")
        self.manager.releaseKey(3)
        self.failUnless(self.manager._acquire_count == 2)
        self.failUnless(self.manager._release_count == 2)
        
    def test_acquire_release_overlapping(self):
        value = self.manager.acquireValue(3)
        self.failUnless(value==9, "Returned the incorrect value.")
        self.failUnless(self.manager._keys_refcount[3] == 1)
        value = self.manager.acquireValue(3)
        self.failUnless(self.manager._keys_refcount[3] == 2)
        self.manager.releaseKey(3)
        self.failUnless(self.manager._keys_refcount[3] == 1)
        self.failUnless(value==9, "Returned the incorrect value.")
        self.manager.releaseKey(3)
        self.failUnless(3 not in self.manager._keys_refcount)
        self.failUnless(self.manager._acquire_count == 1)
        self.failUnless(self.manager._release_count == 2)
     
    def _acquire_key_3(self):
        self.manager.acquireValue(3)
     
    def test_acquire_release_blocking(self):
        cond = threading.Condition()
        self.manager.blockOnAcquire(cond)
        t1 = threading.Thread(target=self._acquire_key_3)
        t2 = threading.Thread(target=self._acquire_key_3)
        self.failUnless(3 not in self.manager._lookup_attempts)
        self.failUnless(3 not in self.manager._keys_refcount)
        t1.start()
        self.failUnless(3 in self.manager._lookup_attempts)
        self.failUnless(3 not in self.manager._keys_refcount)
        t2.start()
        time.sleep(.2)
        cond.acquire()
        cond.notifyAll()
        cond.release()
        cond.acquire()
        cond.notifyAll()
        cond.release()
        t1.join()
        t2.join()
        self.failUnless(self.manager._acquire_count == 1)
        self.failUnless(self.manager._keys_refcount[3] == 2)
        self.manager.releaseKey(3)
        self.manager.releaseKey(3)
        self.failUnless(3 not in self.manager._keys_refcount)
        
    def test_error(self):
        self.manager.errorOnAcquire(Exception("known error."))
        did_fail = False
        try:
            self.manager.acquireValue()
        except:
            did_fail = True
        self.failUnless(did_fail, msg="Error on acquire failed.")
        self.manager.errorOnAcquire(None)
        self.failUnless(self.manager.acquireValue(3) == 9)
        self.failUnless(self.manager._acquire_count == 1)
        self.failUnless(self.manager._keys_refcount[3] == 1)
        self.manager.releaseKey(3)
        self.failUnless(3 not in self.manager._keys_refcount)

if __name__ == '__main__':
    unittest.main()

